/* instruction.c */

/* Copyright (C) 2006 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>

#include "instruction.h"
#include "io.h"
#include "program.h"
#include "record.h"
#include "xalloc.h"

static void
instruction_init (struct instruction *self,
		  enum keyword operation,
		  enum instruction_type type,
		  unsigned lineno)
{
  self->next = self->prev = NULL;
  self->address = 0;
  self->size = 0;
  self->operation = operation;
  self->type = type;
  self->lineno = lineno;
}

struct instruction_binary
{
  struct instruction super;

  const struct symbol *arg[3];
};

struct instruction *
instruction_binary_create(struct program *program, enum keyword op,
			  const struct symbol *dst,
			  const struct symbol *s0, const struct symbol *s1,
			  unsigned lineno)
{
  struct clump *clump = program->clump_end;

  if (!clump)
    return NULL;
    
  if (environment_resolve (clump->local, dst, binding_data, lineno)
      && environment_resolve (clump->local, s0, binding_data, lineno)
      && environment_resolve (clump->local, s1, binding_data, lineno))
    {
      NEW (struct instruction_binary, self);
      instruction_init (&self->super, op, instruction_binary, lineno);

      self->arg[0] = dst;
      self->arg[1] = s0;
      self->arg[2] = s1;

      /* For simplicity, always use long encoding */
      self->super.size = 4;

      return &self->super;
    }
  else
    return NULL;
}

struct instruction *
instruction_finclump_create (unsigned lineno)
{
  NEW (struct instruction, self);
  instruction_init (self, KEYWORD_FINCLUMP, instruction_finclump, lineno);
  self->size = 3;
  return self;
}

/* The total number of words is given by size */
static int
instruction_output_long (FILE *f, enum keyword opcode, unsigned flags, unsigned size)
{
  /* Note: Always uses the long encoding. */
  
  assert (size < 8);
  assert (flags < 8);
  assert (opcode < 0x100);

  /* The bit numbering of the encoding figure on page 25 is
     misleading. From the later example, the opcode byte comes
     first. */
  return (write_uint8 (f, opcode)
	  && write_uint8 (f, (size << 5) | flags));
}

int
instruction_output (struct instruction *self,
		    struct environment *local,
		    struct environment *global,
		    FILE *f)
{
  switch (self->type)
    {
    case instruction_binary:
      {
	unsigned i;
	
	struct instruction_binary *binary
	  = (struct instruction_binary *) self;
	
	if (!instruction_output_long (f, self->operation, 0, self->size))
	  return 0;

	for (i = 0; i < 3; i++)
	  {
	    struct record *record =
	      environment_resolve_record (global, local, binary->arg[i]);

	    if (!write_uint16 (f, record->index))
	      return 0;
	  }
	break;
      }
    case instruction_finclump:
      return (instruction_output_long (f, self->operation, 0, self->size)
	      /* FIXME: No dependants */
	      && write_uint16 (f, 0xffff)
	      && write_uint16 (f, 0xffff));
      
    default:
      abort();
    }
  return 1;
}
