/* nxt-disasseble.c
 *
 */

/* Copyright (C) 2006 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

static void
die (const char *format, ...)
#if __GNUC___
     __attribute__((__format__ (__printf__,1, 2)))
     __attribute__((__noreturn__))
#endif
     ;

static void
die (const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);

  exit(EXIT_FAILURE);
}

static void
werror (const char *format, ...)
#if __GNUC___
     __attribute__((__format__ (__printf__,1, 2)))
#endif
     ;

static void
werror (const char *format, ...)
{
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
}

static void *
xalloc (size_t size)
{
  void *p = malloc (size);
  if (size && !p)
    die ("Virtual memory exhausted.\n");

  return p;
}

static void
read_data (FILE *f, void *buf, size_t length)
{
  size_t res = fread (buf, length, 1, f);
  if (res < 1)
    {
      if (ferror(f))
	die("Read error: %s.\n", strerror(errno));
      else
	die("Unexpected end of file.\n");
    }
}

#if 0
static unsigned
read_uint8 (FILE *f)
{
  unsigned char buf;
  read_data (f, &buf, sizeof(buf));
  return buf;
}

static int
read_int8 (FILE *f)
{
  signed char buf;
  read_data (f, &buf, sizeof(buf));
  return buf;
}
#endif

static unsigned
read_uint16 (FILE *f)
{
  unsigned char buf[2];
  read_data (f, buf, sizeof(buf));
  return buf[0] | (buf[1] << 8);
}

#if 0
static int
read_int16 (FILE *f)
{
  int value = read_uint16 (f);
  if (value & 0x8000)
    value -= 0x10000;

  return value;    
}

static unsigned
read_uint32 (FILE *f)
{
  unsigned char buf[4];
  read_data (f, buf, sizeof(buf));
  return buf[0] | (buf[1] << 8) | (buf[2] << 16) | (buf[3] << 24);
}

static int
read_int32 (FILE *f)
{
  int value = read_uint32 (f);

  if (sizeof(value) > 4 && (value & 0x8000))
    value -= 0x10000;

  return value;    
}
#endif

struct file_header
{
  /* Data space header */
  unsigned data_count;
  unsigned initial_size;
  unsigned static_size;
  unsigned default_size;
  unsigned dynamic_default_offset;
  unsigned dynamic_default_size;
  unsigned dva_head;
  unsigned dva_tail;
  unsigned dope_index;

  unsigned clump_count;
  unsigned code_count;
};

enum data_type {
  TYPE_VOID	= 0x0,
  TYPE_BYTE	= 0x1,
  TYPE_SBYTE	= 0x2,
  TYPE_WORD	= 0x3,
  TYPE_SWORD	= 0x4,
  TYPE_LONG	= 0x5,
  TYPE_SLONG	= 0x6,
  TYPE_ARRAY	= 0x7,
  TYPE_CLUSTER	= 0x8,
  TYPE_MUTEX	= 0x9,
};

struct data_record
{
  enum data_type type;
  int zero_init;
  unsigned value;
};

struct default_data
{
  unsigned static_size;
  const unsigned char *static_data;
  unsigned pos;

  unsigned dynamic_size;
  const unsigned char *dynamic_data;
};

struct clump_record
{
  unsigned fire_count;
  unsigned dependant_count;
  unsigned code_start;
  unsigned char *dependant;
};

static void
read_file_header (FILE *f, struct file_header *header)
{
  static const char magic[16] =
    { 'M', 'i', 'n', 'd', 's', 't', 'o', 'r',
      'm', 's', 'N', 'X', 'T',  0,   0,   5
    };
  char buf[16];

  read_data (f, buf, sizeof(buf));
  if (memcmp (buf, magic, sizeof(buf)) != 0)
    die("Not a NXT executable.\n");

  header->data_count = read_uint16 (f);
  header->initial_size = read_uint16 (f);
  header->static_size = read_uint16 (f);
  header->default_size = read_uint16 (f);
  header->dynamic_default_offset = read_uint16 (f);
  header->dynamic_default_size = read_uint16 (f);
  header->dva_head = read_uint16 (f);
  header->dva_tail = read_uint16 (f);
  header->dope_index = read_uint16 (f);
  header->clump_count = read_uint16 (f);
  header->code_count = read_uint16 (f);
}

static struct data_record *
read_data_toc (FILE *f, const struct file_header *header)
{
  struct data_record *toc;
  unsigned i;

  toc = malloc(header->data_count * sizeof(struct data_record));
  for (i = 0; i < header->data_count; i++)
    {
      unsigned char buf[4];
      read_data (f, buf, sizeof(buf));
      unsigned flags;
      
      toc[i].type = buf[0];
      flags = buf[1];
      toc[i].value = buf[2] | (buf[3] << 8);

      if (toc[i].type > TYPE_MUTEX)
	die("Data entry %d specifies invalid type %d.\n",
	    i, toc[i].type);

      if (flags > 1)
	die("Data entry %d specifies invalid flags 0x%x.\n",
	    i, flags);

      toc[i].zero_init = flags;

      if (toc[i].zero_init)
	{
	  if (toc[i].type == TYPE_MUTEX
	      || toc[i].type == TYPE_ARRAY	      
	      || toc[i].type == TYPE_CLUSTER)
	    die ("Data entry %d specifies zero initialization for type %d.\n",
		 i, toc[i].type);
	}
      else
	{  
	  if (toc[i].type == TYPE_VOID)
	      die ("Data entry %d specifies initial data for type VOID.\n", i);
	}
    }

  return toc;
}

static void
read_default_data (FILE *f, struct default_data *data, const struct file_header *header)
{
  unsigned char *p;

  if (header->default_size != header->dynamic_default_offset + header->dynamic_default_size)
    die ("Inconsistent values for the default data sizes.\n");

  data->static_size = header->dynamic_default_offset;
  data->pos = 0;

  data->dynamic_size = header->dynamic_default_size;

  p = xalloc (data->static_size);
  read_data (f, p, data->static_size);

  data->static_data = p;

  p = xalloc (data->dynamic_size);
  read_data (f, p, data->dynamic_size);

  data->dynamic_data = p;

  if (header->default_size & 1)
    {
      /* Align to even address */      
      char buf;
      read_data (f, &buf, sizeof (buf));
    }
}

static struct clump_record *
read_clumps (FILE *f, const struct file_header *header)
{
  struct clump_record *record;
  unsigned sum;
  unsigned i;

  record = xalloc (header->clump_count * sizeof(struct clump_record));

  for (i = 0; i < header->clump_count; i++)
    {
      unsigned char buf[2];
      read_data (f, buf, sizeof(buf));
      record[i].fire_count = buf[0];
      record[i].dependant_count = buf[1];
      record[i].code_start = read_uint16 (f);

      if (record[i].code_start >= header->code_count)
	die ("Invalid code start %x for clump %d.\n", record[i].code_start, i);
    }

  for (sum = i = 0;
       i < header->clump_count;
       sum += record[i++].dependant_count)
    {  
      if (record[i].dependant_count > 0)
	{
	  record[i].dependant = xalloc (record[i].dependant_count);
	  read_data (f, record[i].dependant, record[i].dependant_count);

	}
      else
	record[i].dependant = NULL;
    }
  if (sum & 1)
    {
      /* Align to even address */      
      char buf;
      read_data (f, &buf, sizeof (buf));
    }  
  return record;
}

static const unsigned *
read_code (FILE *f, unsigned code_count)
{  
  unsigned *code = xalloc (code_count * sizeof(unsigned));
  unsigned i;

  for (i = 0; i < code_count; i++)
    code[i] = read_uint16 (f);

  return code;
}

static void
display_scalar (const unsigned char *data, enum data_type type)
{
  switch (type)
    {
    case TYPE_VOID:
    case TYPE_ARRAY:
    case TYPE_CLUSTER:
      abort();

    case TYPE_BYTE:
      {
	unsigned value = data[0];
	printf("0x%x %u", value, value);
	break;
      }

    case TYPE_SBYTE:
      {
	int value = (int8_t) data[0];
	printf("0x%x %d", value, value);
	break;
      }

    case TYPE_WORD:
      {
	int value = data[0] | (data[1] << 8);
	printf("0x%x %d", value, value);
	break;
      }

    case TYPE_SWORD:
      {
	int value = (int16_t) (data[0] | (data[1] << 8));
	printf("0x%x %d", value, value);
	break;
      }

    case TYPE_LONG:
      {
	int value = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
	printf("0x%x %d", value, value);
	break;
      }
      
    case TYPE_SLONG:
      {
	int value = (int32_t) (data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24));
	printf("0x%x %d", value, value);
	break;
      }

    case TYPE_MUTEX:
      {
	unsigned value = data[0] | (data[1] << 8) | (data[2] << 16) | (data[3] << 24);
	printf("0x%x %d", value, value);
	if (value != 0xffffffff)
	  werror ("Bad mutex initialization.\n");
	break;
      }
    }
}

struct type_info {
  const char *name;
  unsigned size;
};

static unsigned
display_data_record (struct default_data *data,
		     const struct file_header *header,
		     const struct data_record *toc, unsigned index,
		     unsigned indent, int dynamic)
{
  const struct type_info type[TYPE_MUTEX + 1] =
    {
      { "VOID", 0 }, { "BYTE", 1 }, { "SBYTE", 1 },
      { "WORD", 2 }, { "SWORD", 2 }, { "LONG", 4 }, { "SLONG", 4 },
      { "ARRAY", 0 }, { "CLUSTER", 0 }, { "MUTEX", 4 }
    };

  const struct type_info *info;

  if (index >= header->data_count)
    die ("Dataspace table of contents ends prematurely.\n");

  assert (toc[index].type <= TYPE_MUTEX);

  info = &type[toc[index].type];
  printf("%04x: %*s%-7s %04x", index, indent, "",
	 info->name, toc[index].value);

  switch (toc[index].type)
    {
    case TYPE_BYTE: case TYPE_SBYTE:
    case TYPE_WORD: case TYPE_SWORD:
    case TYPE_LONG: case TYPE_SLONG:
    case TYPE_MUTEX:
      if (!dynamic)
	{
	  if (toc[index].zero_init)
	    printf(" <zero>");
	  else
	    {
	      if (data->pos + info->size > data->static_size)
		die ("Static default data ends prematurely.\n");
	      printf(" value: ");
	      display_scalar (data->static_data + data->pos, toc[index].type);
	      data->pos += info->size;
	    }
	}
      /* Fall through */
    case TYPE_VOID:
      printf("\n");
      return index + 1;
    case TYPE_CLUSTER:
      {
	unsigned length;
	unsigned i;

	assert (!toc[index].zero_init);

	length = toc[index++].value;
	printf("\n");
	for (i = 0; i < length; i++)
	  index = display_data_record (data, header, toc, index, indent + 1, dynamic);

	return index;
      }
    case TYPE_ARRAY:
      assert (!toc[index].zero_init);

      printf("\n");
      return display_data_record (NULL, header, toc, index + 1, indent + 1, 1);
    }
  abort();
}

/* Reads initial data. For now, dynamic data is just skipped. */
static void
display_data (struct default_data *data,
	      const struct file_header *header,
	      const struct data_record *toc)
{
  unsigned index;

  for (index = 0; index < header->data_count; )
    index = display_data_record (data, header, toc, index, 0, 0);
}

static void
display_clumps (const struct clump_record *record, unsigned clump_count)
{
  unsigned i;
  for (i = 0; i < clump_count; i++)
    {
      printf("%3d %2d %04x",
	     i, record[i].fire_count, record[i].code_start);

      if (record[i].dependant_count)
	{
	  unsigned j;

	  printf(" { %d", record[i].dependant[0]);
	  
	  for (j = 1; j < record[i].dependant_count; j++)
	    printf (", %d", record[i].dependant[j]);

	  printf(" }");
	}
      printf("\n");
    }
}

/* Note that binary, trinary refers to all arguments, not just the
   source arguments. */
enum op_type {
  OP_UNKNOWN, OP_UNARY, OP_BINARY, OP_TRINARY,
  OP_SUBCALL, /* Immediate target */
};

struct op_info {
  const char *name;
  enum op_type type;
};

#define MAX_OPCODE 0x35

static void
display_instruction (unsigned opcode,
		     const unsigned *args, unsigned nargs, unsigned flags)
{
  static const struct op_info info[MAX_OPCODE + 1] =
    {
      { "ADD", OP_TRINARY }, { "SUB", OP_TRINARY }, { "NEG", OP_BINARY }, { "MUL", OP_TRINARY },
      { "DIV", OP_TRINARY }, { "MOD", OP_TRINARY }, { "AND", OP_TRINARY }, { "OR", OP_TRINARY },
      { "XOR", OP_TRINARY }, { "NOT", OP_BINARY }, { "opx0a", OP_UNKNOWN }, { "opx0b", OP_UNKNOWN },
      { "opx0c", OP_UNKNOWN }, { "opx0d", OP_UNKNOWN }, { "opx0e", OP_UNKNOWN }, { "opx0f", OP_UNKNOWN },
      
      { "opx10", OP_UNKNOWN },{ "CMP", OP_UNKNOWN }, { "TST", OP_UNKNOWN }, { "opx13", OP_UNKNOWN },
      { "opx14", OP_UNKNOWN }, { "INDEX", OP_UNKNOWN }, { "REPLACE", OP_UNKNOWN }, { "ARRSIZE", OP_UNKNOWN },
      { "ARRBUILD", OP_UNKNOWN }, { "ARRSUBSET", OP_UNKNOWN }, { "ARRINIT", OP_UNKNOWN }, { "MOV", OP_BINARY },
      { "SET", OP_UNKNOWN }, { "FLATTEN", OP_UNKNOWN }, { "UNFLATTEN", OP_UNKNOWN }, { "NUM2STR", OP_UNKNOWN },

      { "STR2NUM", OP_UNKNOWN }, { "STRCAT", OP_UNKNOWN }, { "STRSUBSET", OP_UNKNOWN }, { "STR2ARR", OP_UNKNOWN },
      { "ARR2STR", OP_UNKNOWN }, { "JMP", OP_UNKNOWN }, { "BRCMP", OP_UNKNOWN }, { "BRTST", OP_UNKNOWN },
      { "SYSCALL", OP_UNKNOWN }, { "STOP", OP_UNKNOWN }, { "FIN", OP_UNKNOWN }, { "FINIMM", OP_UNKNOWN },
      { "ACQUIRE", OP_UNARY }, { "RELEASE", OP_UNARY }, {"SUBCALL", OP_SUBCALL }, { "SUBRET", OP_UNKNOWN },
      
      { "SETIN", OP_UNKNOWN }, { "SETOUT", OP_UNKNOWN }, { "GETIN", OP_UNKNOWN }, { "GETOUT", OP_UNKNOWN },
      { "opx34", OP_UNKNOWN }, { "GETTICK", OP_UNKNOWN }
    };

  if (opcode > MAX_OPCODE)
    {
      printf("??? (opcode = %x)", opcode);
      goto unknown;
    }

  printf ("%s", info[opcode].name);
  switch (info[opcode].type)
    {
    arg_error:
      printf(" arg error ???");
      /* Fall through */
    unknown:
    case OP_UNKNOWN:
      printf(" flags = %x", flags);

      if (nargs > 0)
	{
	  unsigned j;
	  printf(" { %x", args[0]);
	  for (j = 1; j < nargs; j++)
	    printf(", %x", args[j]);
	  printf(" }");
	}
      break;

    case OP_UNARY:
      {
	if (nargs != 1)
	  goto arg_error;

	printf (" 0x%x", args[0]);
	break;
      }
    case OP_BINARY:
      {
	if (nargs != 2)
	  goto arg_error;

	printf (" 0x%x, 0x%x", args[0], args[1]);
	break;
      }
    case OP_TRINARY:
      {
	if (nargs != 3)
	  goto arg_error;

	printf (" 0x%x, 0x%x, 0x%x", args[0], args[1], args[2]);
	break;
      }
    case OP_SUBCALL:
      {
	if (nargs != 2)
	  goto arg_error;

	printf (" %d, 0x%x", args[0], args[1]);
	break;
      }
    }
  return;
}

static void
display_code (const unsigned *code, unsigned code_count)
{
  unsigned i = 0;

  /* When viewed as 16-bit little-endian numbers, the formats for the
   * first opcode are:
   *
   *      1  1   1   1
   *      5  2   1   0   8   7   0
   *    +------+---+-------+-------+
   *    | size | 1 |  op   |  arg  |     Short format
   *    +------+---+-------+-------+
   *
   *      1  1   1   1
   *      5  2   1   0   8   7    0
   *    +------+---+-------+-------+
   *    | size | 0 | flags |  op   |     Long format
   *    +------+---+-------+-------+
   *
   *      1  1   1   1
   *      5  2   1   0   8   7    0
   *    +------+---+-------+-------+
   *    | 1110 | 0 | flags |  op   |     Very long format
   *    +------+---+-------+-------+
   *
   */
  
   while (i < code_count)
    {
      unsigned instr;
      unsigned opcode;
      unsigned flags;

      /* We want the number of argument words (in counted in words, not bytes). */
      unsigned nargs;

      printf("%04x ", i);
      instr = code[i++];
      if (instr & 0x1000)
	{
	bad_length:
	  printf(" ??? %04x", instr);
	  continue;
	}

      nargs = instr >> 13;
      if (!nargs)
	goto bad_length;

      nargs--;

      if (instr & 0x800)
	{
	  const unsigned short_op[4]
	    /* MOV, ACQUIRE, RELEASE, SUBCALL */
	    = { 0x1b, 0x2c, 0x2d, 0x2e };
	  
	  unsigned args[2];
	  
	  flags = instr & 0xff;
	  if (instr & 0x400)
	    goto bad_length;

	  opcode = short_op[(instr >> 8) & 3];
	  
	  switch (nargs)
	    {
	    case 0:
	      args[0] = flags;
	      display_instruction (opcode, args, 1, 0);
	      break;
	    case 1:
	      if (i == code_count)
		die ("Premature end of code space.\n");
	      args[1] = code[i++];
	      args[0] = (args[1] + flags - ((flags & 0x80) ? 0x100 : 0)) & 0xffff;
	      display_instruction (opcode, args, 2, 0);
	      break;
	    }
	}
      else
	{
	  if (nargs == 0x6)
	    {
	      /* The very long format */
	      if (i == code_count)
		die ("Premature end of code space.\n");
	      nargs = code[i++];
	      if ( (nargs & 1) || nargs < 4)
		goto bad_length;

	      nargs = (nargs/2) - 2;
	    }
	  opcode = instr & 0xff;
	  flags = (instr >> 8) & 7;

	  if (i + nargs > code_count)
	    die ("Premature end of code space.\n");

	  display_instruction (opcode, code + i, nargs, flags);
	  i += nargs;
	}
      printf("\n");
    }
}

int
main (int argc, char **argv)
{
  struct file_header header;
  struct data_record *toc;
  struct default_data default_data;
  struct clump_record *clumps;
  const unsigned *code;
  FILE *f;

  if (argc > 1)
    {
      f = fopen (argv[1], "rb");
      if (!f)
	die ("Opening file '%s' failed: %s.\n",
	     argv[1], strerror(errno));
    }
  else
    f = stdin;

  read_file_header(f, &header);

#define DISPLAY(name, field) \
  printf ("%14s: 0x%04x %4d\n", name, header.field, header.field);

  DISPLAY ("data count", data_count);
  DISPLAY ("initial_size", initial_size);
  DISPLAY ("static_size", static_size);
  DISPLAY ("default_size", default_size);
  DISPLAY ("dynamic_offset", dynamic_default_offset);
  DISPLAY ("dynamic_size", dynamic_default_size);
  DISPLAY ("dva_head", dva_head);
  DISPLAY ("dva_tail", dva_tail);
  DISPLAY ("dope_index", dope_index);
  DISPLAY ("clump_count", clump_count);
  DISPLAY ("code_count", code_count);

  toc = read_data_toc (f, &header);

  read_default_data (f, &default_data, &header);
  display_data (&default_data, &header, toc);

  clumps = read_clumps (f, &header);
  display_clumps (clumps, header.clump_count);

  code = read_code (f, header.code_count);

  display_code (code, header.code_count);

  if (fgetc (f) != EOF)
    werror ("Ignoring extra data after code segment.\n");
    
  return EXIT_FAILURE;
}
