/* parse.y
 *
 * Reads a fairly conventional assembler syntax, with pseudoops to
 * layout data.
 *
 * About symbolic labels:
 *
 *   There are two namespaces, onw global, "top-level", and one local.
 *   Clump labels always belong to the global name space. Instruction
 *   labels always belong to the local namespace. Data labels can be
 *   local or global. Labels need not be defined before use.
 *
 *   A data label is global, if it is defined before the first clump,
 *   or if it is defined usiing the .globl modifier. All other data
 *   labels are local.
 *
 *   Within a clump, when a label is used before being defined, it is
 *   assumed to be local. At the end of each clump, any labels that
 *   are still undefined are moved to the global namespace.
 */

/* Copyright (C) 2006 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

%{
  #if HAVE_CONFIG_H
  # include "config.h"
  #endif

  #include <ctype.h>
  #include <stdarg.h>
  #include <string.h>

  #include "instruction.h"
  #include "keyword.h"
  #include "program.h"
  #include "record.h"
  #include "symbol.h"

  struct parse_state
  {
    FILE *file;
    const char *filename;
    unsigned lineno;

    unsigned nerror;
    struct program *program;
  };
  
  static int
  yylex (struct parse_state *state);

  static void
  yyerror (struct parse_state *state, const char *msg);
%}

%union {
  long number;
  enum keyword keyword;
  const struct symbol *symbol;
  struct instruction *instruction;
}

%debug
%parse-param {struct parse_state *state}
%lex-param {struct parse_state *state}

%token TOK_NEWLINE
%token TOK_ERROR

%token <number> TOK_NUMBER
%token <symbol> TOK_SYMBOL

%token TOK_CLUMP
%token <keyword> TOK_SCALAR_TYPE
%token TOK_ARRAY
%token TOK_CLUSTER
%token TOK_MUTEX

%token <keyword> TOK_OP_BINARY
%token TOK_OP_FINCLUMP

%type <symbol> label
%type <instruction> instruction

%%

program: /* Empty */
	| lines
	;

lines: line
	| lines line
	;

line: TOK_NEWLINE
	| instruction TOK_NEWLINE
		{
		  if (!state->program->clump_end)
		    {
		      fprintf (stderr, "Instruction outside of any clump.\n");
		      YYERROR;
		    }
		  program_add_instruction (state->program, $1);		  
		}
	| label instruction TOK_NEWLINE
		{
		  if (!state->program->clump_end)
		    {
		      fprintf (stderr, "Instruction outside of any clump.\n");
		      YYERROR;
		    }
		  program_add_instruction (state->program, $2);
		  if (!program_define_label(state->program, $1, $2, state->lineno - 1))
		    YYERROR;
		}
	| data TOK_NEWLINE
	| clump TOK_NEWLINE
	| error TOK_NEWLINE
        ;

label: TOK_SYMBOL ':' ;

instruction: TOK_OP_BINARY TOK_SYMBOL ',' TOK_SYMBOL ',' TOK_SYMBOL
		{
		  $$ = instruction_binary_create (state->program, $1, $2, $4, $6, state->lineno);
		}
	| TOK_OP_FINCLUMP
		{
		  $$ = instruction_finclump_create (state->lineno);
		}
	;
data: TOK_SCALAR_TYPE TOK_SYMBOL
		{
		  struct record *record = record_scalar_create ($1, 0);

		  if (!record || !program_add_data(state->program,
						   $2, record, state->lineno))
		    YYERROR;
		}
	| TOK_SCALAR_TYPE TOK_SYMBOL '=' TOK_NUMBER
		{
		  struct record *record = record_scalar_create ($1, $4);

		  if (!record || !program_add_data(state->program,
						   $2, record, state->lineno))
		    YYERROR;
		}
	;

clump: TOK_CLUMP TOK_SYMBOL
		{
		  if (program_end_clump (state->program))
		    program_start_clump (state->program, $2, state->lineno);
		  else
		    YYERROR;
		}

%%

static void
yyerror (struct parse_state *state, const char *msg)
{
  fprintf (stderr, "%s:%d: %s.\n", state->filename, state->lineno, msg);
  state->nerror++;
}

static __attribute__ ((__format__ (__printf__, 2, 3))) int
lexical_error (struct parse_state *state, const char *msg, ...)
{
  va_list args;
  
  va_start (args, msg);

  fprintf (stderr, "%s:%d: ", state->filename, state->lineno);
  vfprintf (stderr, msg, args);
  fprintf (stderr, ".\n");

  va_end (args);
  return TOK_ERROR;
}

static int
skip_ws (FILE *f)
{
  int c;

  do
    c = getc(f);
  while (c != '\n' && isspace (c));

  return c;
}

static void
skip_comment (FILE *f)
{
  int c;

  do
    c = getc(f);
  while (c != EOF && c != '\n');
}

static int
parse_number(FILE *f, long *value, int c)
{
  long x = c - '0';
  for (c = getc(f); isdigit(c); c = getc(f))
    x = 10*x + c - '0';

  ungetc(c, f);
  *value = x;

  return TOK_NUMBER;
}

#define MAX_ID_LENGTH 200

static int
parse_identifier(struct parse_state *state, YYSTYPE *value, int c)
{
  unsigned l;
  char buf[MAX_ID_LENGTH];
  const struct keyword_assoc *keyword;

  buf[0] = c;
  l = 1;
  
  for (;;)
    {
      c = getc (state->file);
      if (! (c == '_' || c == '.' ||
	     (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
	     || (c >= '0' && c <= '9')))
	break;

      if (l >= MAX_ID_LENGTH)
	return lexical_error(state, "Too long identifier");

      buf[l++] = c;
    }
  ungetc(c, state->file);

  keyword = keyword_lookup (buf, l);
  if (keyword)
    {
      value->keyword = keyword->id;
      return keyword->type;
    }

  value->symbol = symbol_intern (state->program->symbols, buf, l, state->lineno);
  return TOK_SYMBOL;
}

static int
yylex (struct parse_state *state)
{
  int c;

  c = skip_ws(state->file);

  switch (c)
    {
    case EOF:
      return -1;
    case '#':
      skip_comment(state->file);
      state->lineno++;
      return TOK_NEWLINE;
    case '\n':
      state->lineno++;
      return TOK_NEWLINE;
    case ':':
    case ',':
    case '=':
      return c;
    default:
      if (isdigit(c))
	return parse_number(state->file, &yylval.number, c);
      else if (c == '_' || c == '.'
	       || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
	return parse_identifier(state, &yylval, c);
      else
	return TOK_ERROR;
    }
}

struct program *
program_parse_file (const char *filename, FILE *f)
{
  struct parse_state state;
  int res;
  
  state.file = f;
  state.filename = filename;
  state.lineno = 1;
  state.nerror = 0;

  state.program = program_create();

  res = yyparse (&state);

  if (res == 0 && program_end_clump (state.program))
    return state.program;
  else
    return NULL;
}
