/* program.c */

/* Copyright (C) 2006 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>

#include "instruction.h"
#include "io.h"
#include "program.h"
#include "record.h"
#include "xalloc.h"

struct clump *
clump_create (void)
{
  NEW (struct clump, self);

  self->next = self->prev = NULL;
  self->local = environment_create();

  self->code_start = self->code_end = NULL;

  self->start = NULL;

  return self;
}

struct program *
program_create (void)
{
  NEW (struct program, self);

  self->symbols = symbol_table_create();
  self->global = environment_create();

  self->data_start = self->data_end = NULL;
  self->clump_start = self->clump_end = NULL;

  return self;
}

#define ADD_TAIL(head, tail, node) do {		\
  (node)->prev = (tail);			\
  (node)->next = NULL;				\
						\
  if ((tail))					\
    (tail)->next = (node);			\
  else						\
    (head) = (node);				\
						\
  (tail) = (node);				\
} while (0)

struct clump *
program_start_clump (struct program *self,
		     const struct symbol *start, unsigned lineno)
{
  struct clump *clump = clump_create ();

  ADD_TAIL(self->clump_start, self->clump_end, clump);

  clump->start = start;
  environment_resolve (clump->local, start, binding_label, lineno);
  
  return clump;
}

int
program_end_clump (struct program *self)
{
  struct clump *clump = self->clump_end;

  if (!clump)
    return 1;

  return environment_handle_undefined (clump->local,
				       self->global);
}

int
program_add_data (struct program *self,
		  const struct symbol *symbol, struct record *record,
		  unsigned lineno)
{
  struct environment *environment
    = (self->clump_end ? self->clump_end->local : self->global);

  if (environment_bind_data (environment, symbol, record, lineno))
    {
      ADD_TAIL (self->data_start, self->data_end, record);
      return 1;
    }
  else
    return 0;
}

void
program_add_instruction (struct program *self, struct instruction *instr)
{
  struct clump *clump = self->clump_end;

  assert (clump);

  ADD_TAIL (clump->code_start, clump->code_end, instr);
}

/* FIXME: Inline in parse.y */
int
program_define_label (struct program *self, const struct symbol *label,
		      struct instruction *definition, unsigned lineno)
{
  return environment_bind_label (self->clump_end->local, label,
				definition, lineno);
}

/* FIXME: Alignment requirements should be stored with the instruction. */
static unsigned
align_address (unsigned address, unsigned align)
{
  unsigned mask = align - 1;

  /* Must be zero or power of two. */
  assert (!(align & mask));
  
  if (align && (address & mask))
    address = (address | mask) + 1;

  return address;
}

void
program_allocate_data (struct program *self)
{
  unsigned location = 0;
  unsigned index = 0;
  unsigned initializer_size = 0;
  struct record *p;

  for (p = self->data_start; p; p = p->next)
    {
      location = align_address (location, p->align);

      p->address = location;
      location += p->size;

      p->index = index;
      index++;

      if (p->value != 0)
	/* Note: No alignment requirements */
	initializer_size += p->size;
    }

  /* Padding for the dope vector array at the start of the dynamic
     data segmenet. FIXME: The example program in the specification
     uses an alignment of 4, is that necessary? */
  location = align_address (location, 4);

  self->data_count = index;
  self->static_data_size = location;
  self->static_initializer_size = initializer_size;
}

void
program_allocate_instructions (struct program *self)
{
  unsigned address = 0;
  unsigned clump_count = 0;
  struct clump *clump;

  for (clump = self->clump_start; clump; clump = clump->next, clump_count ++)
    {
      struct instruction *instr;
      for (instr = clump->code_start; instr; instr = instr->next)
	{
	  instr->address = address;
	  address += instr->size;
	}
    }
  self->codeword_count = address;
  self->clump_count = clump_count;
}

static int
program_output_file_header (struct program *self, FILE *f)
{
  static const char magic[16] =
    { 'M', 'i', 'n', 'd', 's', 't', 'o', 'r',
      'm', 's', 'N', 'X', 'T',  0,   0,   5
    };

  return (WRITE (f, magic, sizeof(magic))

	  /* Note: We don't support any dynamic data yet. That means that the
	     dynamic data section consists of a dope vector array with a
	     single element, of size 10 bytes. */

	  /* The dataspace header */
	  && write_uint16 (f, self->data_count)
	  && write_uint16 (f, self->static_data_size + 10)
	  && write_uint16 (f, self->static_data_size)
	  && write_uint16 (f, self->static_initializer_size + 10)
	  && write_uint16 (f, self->static_initializer_size)
	  && write_uint16 (f, 10)	/* Dynamic initializer size */
	  && write_uint16 (f, 0)	/* DVA head */
	  && write_uint16 (f, 0)	/* and tail */
	  && write_uint16 (f, self->static_data_size)

	  && write_uint16 (f, self->clump_count)
	  && write_uint16 (f, self->codeword_count));
}

static int
program_output_dataspace (struct program *self, FILE *f)
{
  struct record *p;

  /* Write the table of contents */
  for (p = self->data_start; p; p = p->next)
    {
      unsigned type = p->type & 0xff;
      unsigned flags = (p->value == 0);

      assert (type < 16);
      if (! (write_uint8 (f, type) /* The bit numbering in the DSTOC figure on page 11
				    * is misleading. From the later example, the type
				    * byte comes first. */
	     && write_uint8 (f, flags)
	     && write_uint16 (f, p->address)))
	return 0;
    }

  /* Write initial values */
  for (p = self->data_start; p; p = p->next)
    if (p->value)
      {
	switch (p->size)
	  {
	  case 1:
	    if (!write_uint8 (f, p->value))
	      return 0;
	    break;

	  case 2:
	    if (!write_uint16 (f, p->value))
	      return 0;
	    break;

	  case 4:
	    if (!write_uint32 (f, p->value))
	      return 0;
	    break;

	  default:
	    abort();
	  }
      }

  /* Write the dope vector array */
  if (! (write_uint16 (f, self->static_data_size)
	 && write_uint16 (f, 10)	/* Element size */
	 && write_uint16 (f, 1)		/* Current element count */
	 && write_uint16 (f, 0xffff)    /* Unused back pointer */
	 && write_uint16 (f, 0xffff)))	/* Link index */
    return 0;

  if (self->static_initializer_size & 1)
    /* Add a padding byte */
    return write_uint8 (f, 0);

  return 1;
}

static int
program_output_clump_info (struct program *self, FILE *f)
{
  struct clump *clump;

  for (clump = self->clump_start; clump; clump = clump->next)
    {
      struct instruction *start
	= environment_resolve_label (clump->local, clump->start);
      assert (start);

      if (! (write_uint8 (f, 0)		/* Fire count */
	     && write_uint8 (f, 0)	/* Dependant count */
	     && write_uint16 (f, start->address)))
	return 0;
    }

  return 1;
}

static int
program_output_codespace (struct program *self, FILE *f)
{
  struct clump *clump;

  for (clump = self->clump_start; clump; clump = clump->next)
    {
      struct instruction *instr;
      for (instr = clump->code_start; instr; instr = instr->next)
	if (!instruction_output (instr, clump->local, self->global, f))
	  return 0;
    }

  return 1;
}

int
program_output (struct program *self, FILE *f)
{
  return (program_output_file_header (self, f)
	  && program_output_dataspace (self, f)
	  && program_output_clump_info (self, f)
	  && program_output_codespace (self, f));
}
