/* symbol.c */

/* Copyright (C) 2006 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>

#include "symbol.h"
#include "xalloc.h"


/* For simplicity, use a fixed size for the hash tables. In the common
   case, the cluster local table and the global table should be fairly
   small. */
#define SYMBOL_TABLE_SIZE 211
#define ENVIRONMENT_SIZE 211

/* We will use one table for global symbols, and one table of local
   symbols for each cluster. */
struct symbol_table
{
  const struct symbol *table[SYMBOL_TABLE_SIZE];
};


struct symbol_table*
symbol_table_create (void)
{
  NEW (struct symbol_table, self);
  unsigned i;

  for (i = 0; i < SYMBOL_TABLE_SIZE; i++)
    self->table[i] = NULL;

  return self;
}

static unsigned
symbol_hash (const char *name, unsigned length)
{
  unsigned x = length * 1048583;
  unsigned i;

  for (i = 0; i < length; i++)
    x = (x << 5 ) ^ (unsigned char) name[i];

  return x;
}

static const struct symbol *
symbol_create (unsigned hash,
	       const char *name, unsigned name_length,
	       unsigned lineno,
	       const struct symbol *next)
{
  NEW (struct symbol, self);
  self->name = xstrndup (name, name_length);
  self->name_length = name_length;

  self->first_seen_lineno = lineno;

  self->hash_value = hash;
  self->next = next;

  return self;
}

const struct symbol *
symbol_intern (struct symbol_table *self,
	       const char *name, unsigned name_length,
	       unsigned lineno)
{
  unsigned hash = symbol_hash (name, name_length);
  unsigned index = hash % SYMBOL_TABLE_SIZE;
  
  const struct symbol *p;

  for (p = self->table[index]; p; p = p->next)
    if (p->hash_value == hash
	&& p->name_length == name_length
	&& memcmp (name, p->name, name_length) == 0)
      return p;

  p = symbol_create (hash, name, name_length, lineno,
		     self->table[index]);
  self->table[index] = p;
  return p;
}

struct environment {
  struct binding *table[SYMBOL_TABLE_SIZE];
};

struct environment *
environment_create (void)
{
  NEW (struct environment, self);
  unsigned i;

  for (i = 0; i < ENVIRONMENT_SIZE; i++)
    self->table[i] = NULL;

  return self;
}

static struct binding *
binding_create (const struct symbol *symbol, enum binding_type type,
		unsigned lineno)
{
  NEW (struct binding, self);
  self->symbol = symbol;

  self->first_seen_lineno = lineno;

  self->type = type;
  self->definition_lineno = 0;

  return self;
}

static void
binding_destroy (struct binding *self)
{
  free (self);
}

static struct binding *
environment_lookup (struct environment *self,
		    const struct symbol *symbol)
{
  unsigned index = symbol->hash_value % ENVIRONMENT_SIZE;
  
  struct binding *p;

  for (p = self->table[index]; p; p = p->next)
    if (p->symbol == symbol)
      return p;

  return NULL;
}

static void
environment_insert (struct environment *self,
		    struct binding *binding)
{
  unsigned index = binding->symbol->hash_value % ENVIRONMENT_SIZE;

  binding->next = self->table[index];
  self->table[index] = binding;
}

struct binding *
environment_resolve (struct environment *self,
		     const struct symbol *symbol,
		     enum binding_type type,
		     unsigned lineno)
{
  struct binding *p = environment_lookup (self, symbol);

  if (p)
    return (p->type == type) ? p : NULL;

  p = binding_create (symbol, type, lineno);
  environment_insert (self, p);

  return p;
}

struct instruction *
environment_resolve_label (struct environment *self,
			   const struct symbol *symbol)
{
  struct binding *p = environment_lookup (self, symbol);

  if (p && p->type == binding_label && p->definition_lineno)
    return p->definition.instruction;

  else
    return NULL;
}

struct record *
environment_resolve_record (struct environment *self,
			    struct environment *global,
			    const struct symbol *symbol)
{
  struct binding *p = environment_lookup (self, symbol);

  if (!p)
    p = environment_lookup (global, symbol);
  
  if (p && p->type == binding_data && p->definition_lineno)
    return p->definition.record;

  else
    return NULL;
}

static struct binding *
environment_get_binding (struct environment *self,
			 const struct symbol *symbol,
			 enum binding_type type, unsigned lineno)
{
  struct binding *p = environment_lookup (self, symbol);

  if (p)
    {
      if (p->type != type)
	{
	  fprintf (stderr, "Definition of '%s' at line %d inconsistent with use at line %d.\n",
		   symbol->name, lineno, p->first_seen_lineno);
	  return NULL;
	}
      if (p->definition_lineno)
	{
	  fprintf (stderr, "Label '%s' multiply defined, at line %d and line %d.\n",
		   symbol->name, lineno, p->first_seen_lineno);
	  return NULL;
	}
    }
  else
    {
      p = binding_create (symbol, type, lineno);
      environment_insert (self, p);
    }

  return p;
}

int
environment_bind_label (struct environment *self,
			const struct symbol *symbol,
			struct instruction *definition, unsigned lineno)
{
  struct binding *p = environment_get_binding (self, symbol, binding_label, lineno);

  if (!p)
    return 0;

  p->definition.instruction = definition;
  p->definition_lineno = lineno;

  return 1;
}

int
environment_bind_data (struct environment *self,
		       const struct symbol *symbol,
		       struct record *definition, unsigned lineno)
{
  struct binding *p = environment_get_binding (self, symbol, binding_data, lineno);

  if (!p)
    return 0;

  p->definition.record = definition;
  p->definition_lineno = lineno;

  return 1;
}

int
environment_bind_clump (struct environment *self,
			const struct symbol *symbol,
			struct clump *definition, unsigned lineno)
{
  struct binding *p = environment_get_binding (self, symbol, binding_clump, lineno);

  if (!p)
    return 0;

  p->definition.clump = definition;
  p->definition_lineno = lineno;

  return 1;
}

/* Find all undefined local symbols. Data symbols are moved to the
 * global namespace, while undefined code symbols are errors. */
int
environment_handle_undefined (struct environment *self, struct environment *other)
{
  unsigned i;

  for (i = 0; i < ENVIRONMENT_SIZE; i++)
    {
      struct binding **pp;
      struct binding *p;

      for (pp = &self->table[i]; (p = *pp); )
	{
	  if (p->definition_lineno)
	    pp = &p->next;
	  else
	    {
	      /* An undefined symbol. */
	      if (p->type != binding_data || !other)
		{
		  fprintf(stderr, "Missing definition of label '%s', used on line %d.\n",
			  p->symbol->name, p->first_seen_lineno);
		  return 0;
		}
	      else
		{
		  struct binding *global;

		  /* Unlink binding */
		  *pp = p->next;

		  global = environment_lookup (other, p->symbol);
		  if (!global)
		    environment_insert (other, p);
		  else
		    {
		      if (p->type != global->type)
			{
			  fprintf(stderr, "Inconsisten use of global symbol '%s', "
				  "at line %d and %d.\n",
				  p->symbol->name,
				  p->first_seen_lineno,
				  global->first_seen_lineno);
			  return 0;
			}
		      binding_destroy (p);
		    }
		}
	    }
	}
    }
  return 1;
}

void
environment_foreach (struct environment *self,
		     void (*f)(struct binding *, void *),
		     void *ctx)
{
  unsigned i;

  for (i = 0; i < ENVIRONMENT_SIZE; i++)
    {
      struct binding *p;
      for (p = self->table[i]; p; p = p->next)
	f(p, ctx);
    }
}
